"""Common configuration for the atnlp package

"""
import os

# globals
SRC_DIR = os.path.abspath(os.path.dirname(__file__))
TOP_DIR = os.path.abspath(os.path.join(SRC_DIR, '..'))
DATA_DIR = os.path.abspath(os.path.join(TOP_DIR,'data'))
RAW_DATA_DIR = os.path.abspath(os.path.join(DATA_DIR, 'raw'))
EXT_DATA_DIR = os.path.abspath(os.path.join(DATA_DIR, 'external'))
INT_DATA_DIR = os.path.abspath(os.path.join(DATA_DIR, 'interim'))
PRO_DATA_DIR = os.path.abspath(os.path.join(DATA_DIR, 'processed'))
NLTK_DIR = os.path.abspath(os.path.join(RAW_DATA_DIR, 'nltk'))

# set environment variable
os.environ["NLTK_DATA"] = NLTK_DIR



# encoding: utf-8
"""
keras.py
~~~~~~~~

Keras specific evaluation functions

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-06-08"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports 

# third party imports
from keras.callbacks import Callback
import numpy as np
from sklearn.metrics import get_scorer

# local imports

# globals


class Metric(Callback):
    """Class for calculating sklearn metrics on validation data in keras

    :param scorer: sklearn scorer or name of scorer
    :param name: name of metric
    """
    def __init__(self, scorer=None, name=None):
        Callback.__init__(self)
        self.scorer = get_scorer(scorer)
        self.name = name

    def on_epoch_end(self, epoch, logs={}):
        y_true = self.validation_data[1]
        y_pred = (np.asarray(self.model.predict(self.validation_data[0]))).round()
        logs[self.name] = self.scorer._score_func(y_true, y_pred, **self.scorer._kwargs)


# global metrics
f1_metric = Metric(scorer='f1_micro', name='val_f1')

# EOF


#!/usr/bin/env python
# encoding: utf-8
"""
predict.py
~~~~~~~~~~

Generate topic label predictions for input data from trained model

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-06-06"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports
import os
from argparse import ArgumentParser
import logging

# third party imports
import pandas as pd
from sklearn.externals import joblib

# local imports
from atnlp.core.setup import setup
from atnlp.core.logger import log, section_break
from atnlp.core.helpers import start_timer, stop_timer
from atnlp.data.io import read_raw, write_one_hot_labels

# globals


# Your awesome code goes in here...

def build_parser():

    description = "Predict topic labels"
    parser = ArgumentParser(description=description)
    parser.add_argument('model', help="Model config (pickle format)")
    parser.add_argument('data', help="Training data (text file format)")
    parser.add_argument('-o', '--output', default='pred_labels.txt',
                        help="Predicted labels [default: pred_labels.txt]")
    parser.add_argument('-v', '--verbose', action='store_true',
                        help="Set logging level to DEBUG")
    parser.add_argument('-q', '--quiet', action='store_true',
                        help="Quiet mode")

    return parser


def main():

    ti = start_timer()

    # parse command line args
    parser = build_parser()
    args = parser.parse_args()

    # setup atnlp framework
    log_level = logging.DEBUG if args.verbose else None
    if args.quiet: log_level = logging.WARN
    else:
        print("\nExecuting {}\n".format(os.path.basename(__file__)))
    setup(log_level=log_level)

    section_break("Config summary")
    for (k,v) in vars(args).items():
        log().info("{:20s}: {}".format(k,v))

    # ------------------
    # Prepare input data
    # ------------------
    section_break("Preparing input data")

    log().info("Reading data from {}...".format(args.data))
    X = read_raw(args.data)

    # ----------
    # Load model
    # ----------
    section_break("Loading model")

    model = joblib.load(args.model)

    if not args.quiet: print(model)

    # -------
    # Predict
    # -------
    section_break("Predicting labels")
    Y_pred = model.predict(X)
    Y_pred = pd.DataFrame(Y_pred, columns=model.topics)

    # --------
    # Finalize
    # --------
    section_break("Finalizing")
    log().info("Writing labels to {}".format(args.output))
    write_one_hot_labels(Y_pred, args.output)

    # timer
    stop_timer(ti)


if __name__ == "__main__":
    main()
# EOF


# encoding: utf-8
"""Core functionality

.. automodule:: atnlp.core.helpers
    :members:
.. automodule:: atnlp.core.logger
    :members:
.. automodule:: atnlp.core.setup
    :members:

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-05-29"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# EOF


"""Model evaluation

.. automodule:: atnlp.eval.html
    :members:
.. automodule:: atnlp.eval.keras
    :members:
.. automodule:: atnlp.eval.metrics
    :members:
.. automodule:: atnlp.eval.plot
    :members:
.. automodule:: atnlp.eval.table
    :members:

"""


# encoding: utf-8
"""
make_grid.py
~~~~~~~~~~~~~~~

NOTE: work in progress

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-06-01"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports 

# third party imports

# local imports
from atnlp.model import grid

# globals

basepars = {

}

# To check
# - max doc length
# - max vocab size
# - embeddings
# - lstm size / depth
# - dropout
# - bidirectional
# - batch size
# - start with high num epochs

# Strategy
# - start with default setup
# - find optimum epochs (start large)
# - scan batch size
# - scan doc length
# - try min/max length points on bidirectional
# - scan vocab size
# - try dropout
# - try increased depth
# - try more complex embeddings


params = {
    'param1': 1,
    'param2': ['foo', 'bar']
    }



grid.tofile(grid.create(params),
            "-o model.{}.h5",
            "train_reuters_rnn.py",
            "test_grid.sh",
            )


# EOF


"""Model building, training, tuning

.. automodule:: atnlp.model.embed
    :members:
.. automodule:: atnlp.model.grid
    :members:
.. automodule:: atnlp.model.io
    :members:
.. automodule:: atnlp.model.tune
    :members:
.. automodule:: atnlp.model.wordmatch
    :members:

"""


# encoding: utf-8
"""
io.py
~~~~~

Functionality for reading and writing models

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-06-07"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports
import os
import pkg_resources
import yaml

# third party imports
from sklearn.base import clone

# local imports
from atnlp.core.helpers import dynamic_import
from atnlp.core.logger import log

# globals


def create_model(model_name):
    MODEL_DIR = pkg_resources.resource_filename('atnlp', 'share/models/')
    paths = [
        model_name,
        os.path.join(MODEL_DIR, model_name),
        os.path.join(MODEL_DIR, model_name + ".py"),
    ]

    m = None
    for path in paths:
        if os.path.exists(path):
            try:
                mod = dynamic_import(path)
                if hasattr(mod, 'model'):
                    m = mod.model
                else:
                    log().warn("Model configuration script doesn't contain 'model' object")
                if m: break
            except:
                pass
    if not m:
        raise FileNotFoundError("couldn't load model")

    return clone(m)


def load_configured_model(cfg_name):

    CFG_DIR = pkg_resources.resource_filename('atnlp', 'share/config/')
    paths = [
        cfg_name,
        os.path.join(CFG_DIR, cfg_name),
        os.path.join(CFG_DIR, cfg_name + ".yml"),
    ]

    model_name = params = None
    for path in paths:
        if os.path.exists(path):
            try:
                with open(path) as f:
                    data = yaml.load(f)
                    model_name = data['model']
                    params = data['params']
                log().info("loaded model: {}".format(path))
                if model_name: break
            except:
                log.warn("failure parsing config file: {}".format(path))

    if not model_name:
        raise FileNotFoundError("couldn't load model configuration")

    model = create_model(model_name)
    if params:
        model.set_params(**params)

    return model

# EOF


# encoding: utf-8
"""
embed.py
~~~~~~~~

Functionality to load and implement word embeddings (eg. word2vec/glove).

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-05-30"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports 
import os

# third party imports
import gensim
import numpy as np
from keras.layers import Embedding

# local imports
from atnlp.core.logger import log
from atnlp import EXT_DATA_DIR

# globals
EMB_DIR = os.path.join(EXT_DATA_DIR, "embeddings")


def load_glove(filename='glove.6B.300d.w2vformat.txt'):
    """Return glove word embedding model

    The embedding input can be specified with *filename*.
    The inputs are searched for in EMB_DIR.
    Check `scripts/install_glove.py` for installation.

    :param filename: glove input file name
    :return: word embedding model (gensim format)
    """
    filepath = os.path.join(EMB_DIR, filename)
    if not os.path.exists(filepath):
        log().error("failed to load glove embeddings, install with 'install_glove.py'")
        raise FileNotFoundError
    return gensim.models.KeyedVectors.load_word2vec_format(filepath, binary=False)


def create_embedding_layer(w2v, word_to_id, input_length):
    """Return keras embedding layer from pre-trained word embeddings

    :param w2v: pretrained word embedding model
    :param word_to_id: word-to-id dictionary
    :param input_length: length of input (ie number of words) for embedding layer
    :return: keras embedding layer
    """
    vocab_size = len(word_to_id)
    embedding_size = w2v.vector_size
    embedding_matrix = np.zeros((vocab_size, embedding_size))
    for (word, i) in word_to_id.items():
        if word in w2v:
            embedding_matrix[i] = w2v[word]
        else:
            embedding_matrix[i] = np.random.uniform(-0.25,0.25,embedding_size)

    return Embedding(vocab_size, embedding_size,
                     weights=[embedding_matrix],
                     input_length=input_length,
                     trainable=False)


# EOF


#!/usr/bin/env python
# encoding: utf-8
"""
summary.py
~~~~~~~~~~~~~~~

Print performance from model pickle

NOTE: perhaps should be deprecated since it requires specific pickle format that we're not using anymore.

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-06-01"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports 
from argparse import ArgumentParser
import pickle


def main():

    parser = ArgumentParser(description="Print training summary")
    parser.add_argument('file', help="Model training pickle")
    args = parser.parse_args()

    with open(args.file, 'rb') as f:
        scores = pickle.load(f)['scores']
    train, val, test = scores['train'], scores['val'], scores['test']

    print("{:<10s}{:>15s}{:>15s}{:>15s}".format("Sample", "Precision", "Recall", "F1"))
    print("-"*55)
    print("{:<10s}{:15.3f}{:15.3f}{:15.3f}".format("Train", *train[:3]))
    print("{:<10s}{:15.3f}{:15.3f}{:15.3f}".format("Val",   *val[:3]))
    print("{:<10s}{:15.3f}{:15.3f}{:15.3f}".format("Test",  *test[:3]))
    print("")


if __name__ == "__main__":
    main()

# EOF


#!/usr/bin/env python
# encoding: utf-8
"""
train.py
~~~~~~~~

Train topic labelling model on standardized text file input

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-06-06"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports
import os
from argparse import ArgumentParser
import logging

# third party imports
from sklearn.externals import joblib

# local imports
from atnlp.core.setup import setup
from atnlp.core.logger import log, section_break
from atnlp.core.helpers import start_timer, stop_timer
from atnlp.data.io import read_raw, read_one_hot_labels
from atnlp.model.io import load_configured_model

# globals


def build_parser():

    description = "Train topic labelling model"
    parser = ArgumentParser(description=description)
    parser.add_argument('data', help="Training data (text file format)")
    parser.add_argument('labels', help="Training labels (text file format)")
    parser.add_argument('-o', '--output',
                        help="Model output filename [default: {model}.pkl]")
    parser.add_argument('-m', '--model', default='svm',
                        help="Model name or yml config [default: svm]")
    parser.add_argument('-v', '--verbose', action='store_true',
                        help="Set logging level to DEBUG")
    parser.add_argument('-q', '--quiet', action='store_true',
                        help="Quiet mode")

    return parser

def main():

    ti = start_timer()

    # parse command line args
    parser = build_parser()
    args = parser.parse_args()
    if not args.output:
        model_name = os.path.splitext(os.path.basename(args.model))[0]
        if model_name:
            args.output = model_name + '.pkl'
        else:
            args.output = 'model.pkl'

    # setup atnlp framework
    log_level = logging.DEBUG if args.verbose else None
    if args.quiet: log_level = logging.WARN
    else:
        print("\nExecuting {}\n".format(os.path.basename(__file__)))
    setup(log_level=log_level)

    section_break("Config summary")
    for (k,v) in vars(args).items():
        log().info("{:20s}: {}".format(k,v))

    # ------------------
    # Prepare input data
    # ------------------
    section_break("Preparing input data")

    log().info("Reading training data from {}...".format(args.data))
    X = read_raw(args.data)

    log().info("Reading training labels from {}...".format(args.labels))
    Y = read_one_hot_labels(args.labels)

    # ------------
    # Create model
    # ------------
    section_break("Creating model")

    # dynamically load model using yml config
    model = load_configured_model(args.model)

    # attach topics to model so they are persistified
    model.topics = list(Y.columns)

    if not args.quiet:
        log().info("")
        for s in str(model).split('\n'):
            log().info(s)

    # ---------
    # Fit model
    # ---------
    section_break("Training model")
    model.fit(X, Y)

    # --------
    # Finalize
    # --------
    section_break("Finalizing")
    log().info("Saving model to {}".format(args.output))
    joblib.dump(model, args.output)

    # timer
    stop_timer(ti)


if __name__ == "__main__":
    main()

# EOF


#!/usr/bin/env python
# encoding: utf-8
"""
train_reuters_rnn.py
~~~~~~~~~~~~~~~~~~~~

Train RNN for topic labelling on reuters dataset using keras

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-05-29"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports 
from argparse import ArgumentParser
import logging
import os
import pickle

# third party imports
from keras.preprocessing.sequence import pad_sequences
from keras.models import Sequential
from keras.layers import LSTM, Embedding, Dense, Bidirectional
from keras.callbacks import EarlyStopping
from sklearn.model_selection import train_test_split
from sklearn.metrics import precision_recall_fscore_support

# local imports
from atnlp.core.setup import setup
from atnlp.core.logger import log, section_break, title_break
from atnlp.core.helpers import start_timer, stop_timer
from atnlp.data.reuters import get_data, get_labels, get_topics
from atnlp.data.parse import build_vocab, raw_to_ids, PAD_WORD
from atnlp.model.embed import load_glove, create_embedding_layer
from atnlp.eval.metrics import f1_metric

# globals


def build_parser():

    description = "Train RNNs for reuters dataset"
    parser = ArgumentParser(description=description)
    parser.add_argument('-o', '--output', default='model.h5',
                        help="Model output filename [default: model.h5]")
    parser.add_argument('-v', '--verbose', action='store_true',
                        help="Set logging level to DEBUG")

    proc_group = parser.add_argument_group('Preprocessing configuration')
    proc_group.add_argument('--min-samples', type=int, default=100,
                        help="Minimum number of samples per category [default: 100]")
    proc_group.add_argument('--topics',
                        help="comma separated list of topics")
    proc_group.add_argument('--max-doc-length', type=int, default=100,
                        help="Maximum number of words per document [default: 100]")
    proc_group.add_argument('--max-vocab-size', type=int, default=10000,
                        help="Maximum number of words in vocabulary [default: 10000]")

    model_group = parser.add_argument_group('Model configuration')
    model_group.add_argument('-e', '--learn-embeddings', action="store_true",
                        help="Learn word embeddings from data")
    model_group.add_argument('--embedding-size', type=int, default=32,
                        help="Size of embedding layer if not pretrained [default: 32]")
    model_group.add_argument('-s', '--lstm-size', type=int, default=None,
                        help="Size of LSTM layer [default: use embedding size]")
    model_group.add_argument('-d', '--lstm-depth', type=int, default=1,
                        help="Depth of LSTM layer [default: 1]")
    model_group.add_argument('--bidirectional', action="store_true",
                        help="Use bidirectional LSTM [default: False]")
    model_group.add_argument('--dropout', type=float, default=0.0,
                        help="Dropout fraction [default: 0.0]")
    model_group.add_argument('--recurrent-dropout', type=float, default=0.0,
                        help="Recurrent dropout fraction [default: 0.0]")

    fit_group = parser.add_argument_group('Fit configuration')
    fit_group.add_argument('-b', '--batch-size', type=int, default=5,
                        help="Batch size [default: 5]")
    fit_group.add_argument('-n', '--epochs', type=int, default=10,
                        help="Number of epochs [default: 10]")
    fit_group.add_argument('--no-early-stopping', action='store_true',
                        help="Don't use early stopping")

    return parser

def main():

    ti = start_timer()

    # parse command line args
    parser = build_parser()
    args = parser.parse_args()

    assert args.lstm_depth >= 1, "Must configure at least one LSTM layer"

    print("\nExecuting train_reuters_rnn.py\n")

    # setup atnlp framework
    log_level = logging.DEBUG if args.verbose else None
    setup(log_level=log_level)

    section_break("Config summary")
    for (k,v) in vars(args).items():
        log().info("{:20s}: {}".format(k,v))

    # ------------------
    # Prepare input data
    # ------------------
    section_break("Preparing input data")

    # select topics
    if args.topics:
        topics = args.topics.split(',')
    else:
        topics = get_topics(min_samples=args.min_samples)
    log().info("{} topics selected.".format(len(topics)))

    # get topic labels (MxN data frame of bools: M categories, N documents)
    # TODO: could explicitly ignore Y_test here to show we don't need test labels
    log().info("getting topic labels...")
    (Y_train, Y_test) = get_labels(topics)

    # get data iterators
    # Note: we also use test data because model currently requires
    #       vocab from all samples to get be predictions
    log().info("getting topic data...")
    (X_train_raw, X_test_raw) = get_data(topics)

    # convert words to integers
    log().info("converting to integer representation...")
    word_to_id = build_vocab(list(X_train_raw) + list(X_test_raw), max_size=args.max_vocab_size)
    X_train_ids = raw_to_ids(X_train_raw, word_to_id)
    X_test_ids = raw_to_ids(X_test_raw, word_to_id)

    # pad
    log().info("padding sequences...")
    id_to_word = dict(zip(word_to_id.values(), word_to_id.keys()))
    X_train_ids = pad_sequences(X_train_ids, maxlen=args.max_doc_length, value=word_to_id[PAD_WORD],
                                padding='post', truncating='post')
    X_test_ids = pad_sequences(X_test_ids, maxlen=args.max_doc_length, value=word_to_id[PAD_WORD],
                               padding='post', truncating='post')
    vocab_size = len(word_to_id)

    # split train into train + validation
    X_train_ids, X_val_ids, Y_train, Y_val = train_test_split(
        X_train_ids, Y_train, test_size=0.20, random_state=42)

    # dataset summary
    title_break("Data Summary")
    log().info("{} topics selected: {}".format(len(topics), topics))
    log().info("n train: {}".format(len(X_train_ids)))
    log().info("n val:   {}".format(len(X_val_ids)))
    log().info("n test:  {}".format(len(X_test_ids)))
    log().info("max doc length: {}".format(args.max_doc_length))
    log().info("vocab size: {}".format(vocab_size))

    # ------------
    # Create model
    # ------------
    section_break("Creating Model")
    # create embedding layer
    if args.learn_embeddings:
        embedding = Embedding(vocab_size, args.embedding_size)
    else:
        embedding = create_embedding_layer(load_glove(), word_to_id, args.max_doc_length)

    # create LSTM layers
    lstm_size = args.lstm_size or embedding.output_dim
    lstm_args = {'dropout':args.dropout, 'recurrent_dropout': args.recurrent_dropout}
    lstm_layers = [LSTM(lstm_size, **lstm_args) for _ in range(args.lstm_depth)]
    if args.bidirectional:
        lstm_layers = [Bidirectional(l) for l in lstm_layers]

    # construct model
    model = Sequential()
    model.add(embedding)
    for l in lstm_layers: model.add(l)
    model.add(Dense(units=len(topics), activation='sigmoid'))
    model.compile(loss='binary_crossentropy', optimizer='adam', metrics=['accuracy'])

    print(model.summary())

    # ---------
    # Fit model
    # ---------
    # create callbacks
    callbacks = [f1_metric]
    if not args.no_early_stopping:
        callbacks += [EarlyStopping(monitor='val_f1', mode='max', patience=2)]

    # fit
    section_break("Training Model")
    history = model.fit(X_train_ids, Y_train, epochs=args.epochs,
                       batch_size=args.batch_size, verbose=1,
                       validation_data=(X_val_ids, Y_val),
                       callbacks=callbacks)

    # --------------
    # Evaluate model
    # --------------
    section_break("Evaluating Model")
    threshold = 0.5
    Y_train_pred = model.predict(X_train_ids) > threshold
    Y_val_pred = model.predict(X_val_ids) > threshold
    Y_test_pred = model.predict(X_test_ids) > threshold
    ave = 'micro'
    scores_train = precision_recall_fscore_support(Y_train, Y_train_pred, average=ave)
    scores_val   = precision_recall_fscore_support(Y_val, Y_val_pred, average=ave)
    scores_test  = precision_recall_fscore_support(Y_test, Y_test_pred, average=ave)

    title_break("Performance")
    log().info("{:<10s}{:>15s}{:>15s}{:>15s}".format("Sample", "Precision", "Recall", "F1"))
    log().info("-"*55)
    log().info("{:<10s}{:15.3f}{:15.3f}{:15.3f}".format("Train", *scores_train[:3]))
    log().info("{:<10s}{:15.3f}{:15.3f}{:15.3f}".format("Val",   *scores_val[:3]))
    log().info("{:<10s}{:15.3f}{:15.3f}{:15.3f}".format("Test",  *scores_test[:3]))
    log().info("")

    # timer
    dt = stop_timer(ti)

    # --------
    # Finalize
    # --------
    section_break("Finalizing")
    log().info("Saving model to {}".format(args.output))
    model.save(args.output)
    auxname = os.path.splitext(args.output)[0] + '.pickle'
    log().info("Saving aux info to {}".format(auxname))
    with open(auxname, 'wb') as f:
        data = {
            'topics': topics,
            'id_to_word':id_to_word,
            'history': history.history,
            'scores':{
                'train': scores_train,
                'val': scores_val,
                'test': scores_test,
            },
            'time': dt.total_seconds(),
            'args': vars(args),
        }
        pickle.dump(data, f, pickle.HIGHEST_PROTOCOL)

    # TODO: use cross validation?


if __name__ == "__main__":
    main()

# EOF



# encoding: utf-8
"""
helpers.py
~~~~~~~~~~

Collection of global framework helper functions

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-05-30"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports 
from time import time
from datetime import timedelta
import importlib.util
import pkgutil
import os
from importlib import import_module

# third party imports

# local imports
from atnlp.core.logger import log

# globals


def start_timer():
    """Return current time"""
    return time()


def stop_timer(ti):
    """Summarize job timing"""
    dt = timedelta(seconds=(time()-ti))
    time_str = str(dt)
    log().info("Execution time: {0}".format(time_str))
    return dt


def dynamic_import(path):
    """Import and return module from it's file path

    :param path: path to module in file system
    :return: module
    """
    spec = importlib.util.spec_from_file_location('tmpmod', path)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    return mod


def get_modules(module):
    """Recursively return all submodule names

    :param module: module name
    :return: list of submodules
    """
    m = import_module(module)
    module_path = os.path.dirname(m.__file__)
    paths = []
    for _, name, ispkg in pkgutil.iter_modules([module_path]):
        submod = ".".join([module,name])
        if ispkg: paths += get_modules(submod)
        else:     paths.append(submod)
    return paths

# EOF


#!/usr/bin/env python
# encoding: utf-8
"""
evaluate.py
~~~~~~~~~~~

<Description goes here...>

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-06-06"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports
import logging
import os
from argparse import ArgumentParser

# third party imports
from sklearn.externals import joblib

# local imports
from atnlp.core.helpers import start_timer, stop_timer
from atnlp.core.logger import log, section_break
from atnlp.core.setup import setup
from atnlp.data.io import read_raw, read_one_hot_labels
from atnlp.eval.html import Report
from atnlp.eval.plot import topic_labelling_barchart, topic_correlation_matrix, \
    topic_migration_matrix, false_labels_matrix
from atnlp.eval.table import multimodel_topic_labelling_summary_tables

# globals


def build_parser():
    description = "Evaluate topic models"
    parser = ArgumentParser(description=description)
    parser.add_argument('data', help="Data (text file format)")
    parser.add_argument('labels', help="Labels (text file format)")
    parser.add_argument('models', nargs='+', help="Model configs (pickle format)")
    parser.add_argument('-o', '--output', default='summary.html',
                        help="Directory for evaluation outputs [default: summary]")
    parser.add_argument('-v', '--verbose', action='store_true',
                        help="Set logging level to DEBUG")
    parser.add_argument('-q', '--quiet', action='store_true',
                        help="Quiet mode")

    return parser


def main():
    ti = start_timer()

    # parse command line args
    parser = build_parser()
    args = parser.parse_args()

    # setup atnlp framework
    log_level = logging.DEBUG if args.verbose else None
    if args.quiet:
        log_level = logging.WARN
    else:
        print("\nExecuting {}\n".format(os.path.basename(__file__)))
    setup(log_level=log_level)

    section_break("Config summary")
    for (k, v) in vars(args).items():
        log().info("{:20s}: {}".format(k, v))

    # ------------------
    # Prepare input data
    # ------------------
    section_break("Preparing input data")

    log().info("Reading training data from {}...".format(args.data))
    X = read_raw(args.data)

    log().info("Reading training labels from {}...".format(args.labels))
    Y = read_one_hot_labels(args.labels)

    # -----------
    # Load models
    # -----------
    section_break("Loading models")
    names = [os.path.splitext(m)[0] for m in args.models]
    models = [joblib.load(m) for m in args.models]

    # -------
    # Predict
    # -------
    section_break("Predicting labels")
    preds = [m.predict(X) for m in models]

    # --------
    # Evaluate
    # --------
    tables = multimodel_topic_labelling_summary_tables(Y, preds, names)

    # --------
    # Finalize
    # --------
    section_break("Finalizing")

    html = Report()
    html.add_title("Topic modelling performance",
                   par="Here are some totally awesome results on topic modelling!")

    html.add_section("Topic-averaged performance")
    html.add_text("The precision, recall and f1 metrics use 'micro' averaging over topics")
    html.add_table(tables['summary'], cap='')

    html.add_section("Per-topic performance")

    topic_labelling_barchart(Y, preds, names)
    html.add_figure(cap="Comparison of per-topic metrics for each model")

    html.add_section("Per-topic performance (tables)")

    html.add_table(tables['precision'], cap='Precision scores per topic for each model')
    html.add_table(tables['recall'], cap='Recall scores per topic for each model')
    html.add_table(tables['f1'], cap='f1 scores per topic for each model')
    html.add_table(tables['fl'], cap='Number of false labels')
    html.add_table(tables['ml'], cap='Number of missed labels')

    # best model perf
    best_model = tables['summary']['model'].iloc[0]
    best_index = names.index(best_model)
    best_pred = preds[best_index]

    html.add_section("Correlations")
    topic_correlation_matrix(Y)
    html.add_figure(cap='True topic correlations')

    topic_migration_matrix(Y, best_pred)
    html.add_figure(cap='Topic migration matrix')

    false_labels_matrix(Y, best_pred)
    html.add_figure(cap="False labels matrix")

    html.write(args.output)

    # log().info("Writing labels to {}".format(args.output))

    # timer
    stop_timer(ti)


if __name__ == "__main__":
    main()
# EOF



# encoding: utf-8
"""
html.py
~~~~~~~

Classes for rendering documents as html.

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-06-06"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports 
import base64
from io import BytesIO

# third party imports
from matplotlib import pyplot as plt

# local imports

# globals


class Report(object):
    """Simple html report class based on bootstrap

    Use the interface to add elements (title, figures, tables, text, etc),
    then use *write* to dump rendered html to file.

    """
    def __init__(self):
        self._components = []
        self._nfig = 0
        self._ntab = 0

    def add_title(self, title, par=None):
        """Add title to document

        :param title: title string
        :param par: paragraph to go with title (optional)
        """
        s = """      <div class="jumbotron">\n        <h1>{}</h1>""".format(title)
        if par: s += """        <p>{}</p>""".format(par)
        s += """      </div>"""
        self._add(s)

    def add_figure(self, cap=""):
        """Add figure to document

        Call this directly after creating a figure with matplotlib.
        The figure will be embedded into the html document.

        :param cap: figure caption (optional)
        """
        self._nfig += 1
        image = BytesIO()
        plt.savefig(image, format='png')
        fig = base64.encodebytes(image.getvalue()).decode()
        htmlfig = """<div align="center">
            <p> <b> Figure {index}:</b> {cap}</p>
            </div> 
            <div class="page-header">
            <img src="data:image/png;base64,{fig}" class="img-thumbnail"/>
            </div>
        """.format(index=self._nfig, cap=cap, fig=fig)
        self._add(htmlfig)

    def add_table(self, tab, cap=""):
        """Add table to document

        :param tab: table (pandas DataFrame)
        :param cap: caption (optional)
        """
        self._ntab += 1
        t = tab.to_html(index=False)
        t = t.replace('<table border="1" class="dataframe">',
                      '<table class="table table-striped">')

        htmltab = """<div align="center">
        <p> <b> Table {index}:</b> {cap}</p>
        </div>
        {tab}""".format(index=self._ntab, cap=cap, tab=t)
        self._add(htmltab)

    def add_styled_table(self, tab, cap=""):
        """Add styled table to document

        Note: full control over html style is given to the Styler and bootstrap css
        is not used, so it can be difficult to get something that actually
        looks good.

        :param tab: table (pandas Styler)
        :param cap: caption (optional)
        """
        self._ntab += 1
        htmltab = """<div align="center">
        <p> <b> Table {index}:</b> {cap}</p>
        </div>
        {tab}""".format(index=self._ntab, cap=cap, tab=tab.render())
        self._add(htmltab)


    def add_section(self, title):
        """Add section to document

        :param title: section title
        """
        html =  """
      <div class="page-header">
        <h2>{}</h2>
      </div>        
        """.format(title)
        self._add(html)

    def add_text(self, text):
        """Add paragraph text to document

        :param text: text string
        """
        html = """<p>{}</p>""".format(text)
        self._add(html)

    def write(self, filename):
        """Write rendered html to file

        :param filename: path to output file
        """
        with open(filename, 'w') as f:
            f.write(self._render())

    def _add(self, text):
        """Add an element to docuemnt (internal function)

        :param text: rendered html text
        """
        self._components.append(text)

    def _render(self):
        """Return complete document in rendered html format (internal function)

        :return: rendered html document (string)
        """
        return '\n'.join([self._header()] + self._components + [self._footer()])

    def _header(self):
        """Return html header

        :return: string
        """
        return """
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <!-- The above 3 meta tags *must* come first in the head; any other head content must come *after* these tags -->
    <meta name="description" content="">
    <meta name="author" content="">

    <title>This is a title</title>

    <!-- Bootstrap core CSS -->
    <link href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/css/bootstrap.min.css" rel="stylesheet">
    <!-- Bootstrap theme -->
    <link href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/css/bootstrap-theme.min.css" rel="stylesheet">
    <!-- IE10 viewport hack for Surface/desktop Windows 8 bug -->
    <link href="https://maxcdn.bootstrapcdn.com/css/ie10-viewport-bug-workaround.css" rel="stylesheet">
    <!-- Custom styles for this template -->
    <link href="https://getbootstrap.com/docs/3.3/examples/theme/theme.css" rel="stylesheet">
  </head>

  <body> 
  
  <div class="container theme-showcase" role="main">
        """

    def _footer(self):
        """Return html footer

        :return: string
        """
        return """    </body>\n\n    </html>"""


# EOF


# encoding: utf-8
"""
plot.py
~~~~~~~

Functionality for creating performance summary plots.

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-06-08"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports

# third party imports
import random
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.colors import CSS4_COLORS
from sklearn.metrics import accuracy_score
from sklearn.metrics.scorer import check_scoring, _PredictScorer, _ProbaScorer
from sklearn.base import clone as clone_model
from sklearn.model_selection import cross_val_score

# local imports
from atnlp.eval.table import topic_labelling_summary_table

# globals
COLORS = ['red', 'blue', 'green', 'orange', 'magenta', 'yellow', 'brown']
random.seed(42)
COLORS += random.sample(list(CSS4_COLORS), len(CSS4_COLORS))


def create_awesome_plot_grid(nminor, ncol=5, maj_h=2, maj_w=3,
                             min_xlabel=None, min_ylabel=None,
                             maj_xlabel=None, maj_ylabel=None,
                             grid=True):
    """Returns an awesome plot grid

     The grid includes a specified number (*nminor*) of minor
     plots (unit size in the grid) and a single major plot
     whose size can be specified in grid units (*maj_h* and *maj_w*).

     The major plot is located top-right. If either dimension
     is 0 the major plot is omitted.

     The minor plots are tiled from left-to-right, top-to-bottom
     on a grid of width *ncol* and will be spaced around the major
     plot.

    The grid will look something like this

    .. code-block:: text

        #----#----#----#---------#
        |    |    |    |         |
        |    |    |    |         |
        #----#----#----#         |
        |    |    |    |         |
        |    |    |    |         |
        #----#----#----#----#----#
        |    |    |    |    |    |
        |    |    |    |    |    |
        #----#----#----#----#----#
        |    |    |
        |    |    | -->
        #----#----#


    :param nminor: number of minor plots
    :param ncol: width of grid (in grid units)
    :param maj_h: height of major plot (in grid units)
    :param maj_w: width of major plot (in grid units)
    :param min_xlabel: x-axis label of minor plots
    :param min_ylabel: y-axis label of minor plots
    :param maj_xlabel: x-axis label of major plot
    :param maj_ylabel: y-axis label of major plot
    :param grid: draw grid lines (if True)
    :return: tuple (figure, major axis, minor axes (flat list), minor axes (2D list))
    """
    assert maj_w <= ncol, "Major fig cannot be wider than grid!"

    def pad_coord(ipad):
        """Return x-y coordinate for ith element"""
        i = int(np.floor(ipad / ncol))
        j = ipad % ncol
        return (i, j)

    def in_main(ipad):
        """Return True if ith element within major plot space"""
        (i, j) = pad_coord(ipad)
        if j >= ncol - maj_w and i < maj_h: return True
        return False

    # derived quantities
    n = maj_w * maj_h + nminor
    nrow = int(np.ceil(n / ncol))
    if maj_h and nminor <= ncol - maj_w:
        ncol = maj_w + nminor
    if maj_w:
        nrow = max(nrow, maj_h)

    # create figure
    f = plt.figure(figsize=(16 * ncol / 5, 16 * nrow / 5))

    # create major axis
    if maj_h and maj_w:
        ax_maj = plt.subplot2grid((nrow, ncol), (0, ncol - maj_w), colspan=maj_w, rowspan=maj_h)
        if maj_xlabel: ax_maj.set_xlabel(maj_xlabel)
        if maj_ylabel: ax_maj.set_ylabel(maj_ylabel)
        ax_maj.tick_params(top=True, right=True,
                           labeltop=True, labelright=True,
                           labelleft=False, labelbottom=False,
                           grid_linestyle='-.')
        ax_maj.grid(grid)
    else:
        ax_maj = None

    # create minor axes
    ax_min = []
    ax_min_ij = [[None] * ncol] * nrow
    ipad = 0
    imin = 0
    while imin < nminor:
        if not in_main(ipad):
            (i, j) = pad_coord(ipad)
            ax0 = ax_min[0] if ax_min else None
            ax = plt.subplot2grid((nrow, ncol), (i, j), sharex=ax0, sharey=ax0)
            ax.i = i
            ax.j = j
            ax.tick_params(top=True, right=True, grid_linestyle='-.')
            ax.grid(grid)
            # add top labels
            if i == 0:
                ax.tick_params(labeltop=True)

            # add right labels
            if j == ncol - 1: ax.tick_params(labelright=True)
            # remove inner left labels
            if j > 0:
                ax.tick_params(labelleft=False)
            # set y-titles
            elif min_ylabel:
                ax.set_ylabel(min_ylabel)
            # set x-titles
            if min_xlabel: ax.set_xlabel(min_xlabel)
            # remove inner bottom labels
            if i > 0 and ax_min_ij[i - 1][j]:
                ax_min_ij[i - 1][j].tick_params(labelbottom=False)
                ax_min_ij[i - 1][j].set_xlabel("")

            ax_min.append(ax)
            ax_min_ij[i][j] = ax
            imin += 1
        ipad += 1

    return (f, ax_maj, ax_min, ax_min_ij)


def binary_classification_accuracy_overlays(classifiers, X_train, y_train, X_test, y_test):
    """Create overlays of binary classification accuracy for multiple classifiers

    :param classifiers: list of tuples (name, classifier)
    :param X_train: training data
    :param y_train: binary training labels
    :param X_test: testing data
    :param y_test: binary testing labels
    :return: tuple (figure, axis)
    """
    acc_train = [accuracy_score(y_train, c.predict(X_train))
               for (_,c) in classifiers]
    acc_test = [accuracy_score(y_test, c.predict(X_test))
               for (_,c) in classifiers]

    acc_cv = [c.cv_results_['mean_test_score'][c.best_index_]
             for (_,c) in classifiers]
    acc_err_cv = [c.cv_results_['std_test_score'][c.best_index_]
             for (_,c) in classifiers]

    names = [n for (n,_) in classifiers]
    ypos = np.arange(len(classifiers))

    fig, ax = plt.subplots()
    ax.barh(ypos, acc_cv, xerr=acc_err_cv, align='center',
            color='g', label='cv', alpha=0.5)
    ax.set_yticks(ypos)
    ax.set_yticklabels(names)
    ax.set_xlabel('Accuracy')
    ax.scatter(acc_train, ypos, color='red', label='train')
    ax.scatter(acc_test, ypos,  color='b',   label='test')
    ax.invert_yaxis()
    ax.legend()
    xmin = 0.98 * min(acc_train+acc_test+acc_cv)
    xmax = 1.02 * max(acc_train+acc_test+acc_cv)
    ax.set_xlim(xmin,xmax)

    return (fig, ax)


def topic_labelling_scatter_plots(Y_true, Y_pred, sample_min=None, thresholds=None):
    """Create scatter plots comparing precision, recall and number of samples

    :param Y_true: ground truth topic labels (one-hot format)
    :param Y_pred: topic predictions (one-hot format)
    :param sample_min: minimum number of examples per topic
    :param thresholds: list of thresholds per category (optional)
    :return: tuple (figure, list of axes)
    """
    table = topic_labelling_summary_table(Y_true, Y_pred, sample_min, thresholds)

    # Make scatter plots
    f = plt.figure(figsize=(20,5))

    ax1 = plt.subplot(1,3,1)
    ax1.scatter(table['recall'], table['precision'])
    plt.xlabel('recall')
    plt.ylabel('contamination')

    ax2 = plt.subplot(1,3,2)
    ax2.scatter(table['samples'], table['recall'])
    ax2.set_xscale('log')
    plt.xlabel('samples')
    plt.ylabel('recall')

    ax3 = plt.subplot(1,3,3)
    ax3.scatter(table['samples'], table['precision'])
    ax3.set_xscale('log')
    plt.xlabel('samples')
    plt.ylabel('contamination')

    return (f, (ax1, ax2, ax3))


def topic_labelling_barchart(Y_true, Y_preds, model_names):
    """Create topic labelling barchart

    The figure includes a 1x4 grid of bar charts, illustrating
    the number of samples, precision, recall and f1 scores for
    each topic. The scores are overlayed for each model.

    :param Y_true: ground truth topic labels (one-hot format)
    :param Y_preds: topic predictions for each model (list of one-hot formats)
    :param model_names: topic labelling model names
    :return: tuple (figure, list of axes)
    """
    n = len(model_names)
    tables = [topic_labelling_summary_table(Y_true, Y_preds[i]) for i in range(n)]
    topics = tables[0]['topic']
    samples = tables[0]['samples']

    # y-axis
    ypos = np.arange(len(samples))

    # figure
    plt.close('all')
    f, (ax1, ax2, ax3, ax4) = plt.subplots(1, 4, sharey=True, figsize=(16, 0.25 * len(samples)))

    # samples subfig
    ax1.set_xlabel('Samples')
    ax1.barh(ypos, samples, align='center',
             color='g', label='Samples', alpha=0.25)
    ax1.set_yticks(ypos)
    ax1.set_yticklabels(topics)
    ax1.invert_yaxis()

    # precision
    ax2.set_xlabel('Precision')
    ax2.set_xlim((-0.05, 1.05))
    for i in range(n):
        ax2.scatter(tables[i]['precision'], ypos, color=COLORS[i],
                    label=model_names[i], alpha=0.5)

    # recall
    ax3.set_xlabel('Recall')
    ax3.set_xlim((-0.05, 1.05))
    for i in range(n):
        ax3.scatter(tables[i]['recall'], ypos, color=COLORS[i],
                    label=model_names[i], alpha=0.5)

    # recall
    ax4.set_xlabel('F1')
    ax4.set_xlim((-0.05, 1.05))
    for i in range(n):
        ax4.scatter(tables[i]['f1'], ypos, color=COLORS[i],
                    label=model_names[i], alpha=0.5)


    ax4.legend(loc='center left', bbox_to_anchor=(1, 1))

    gridlines = []
    for ax in [ax1, ax2, ax3, ax4]:
        ax.grid()
        gridlines += ax.get_xgridlines() + ax.get_ygridlines()
    for line in gridlines:
        line.set_linestyle('-.')

    return (f, (ax1, ax2, ax3, ax4))


def topic_labelling_barchart_cv(models, model_names, model_inputs, Y, cv=10):
    """Create topic labelling barchart with k-fold cross-validation

    Figure layout is the same as in :func:`topic_labelling_barchart`.

    K-fold cross-validation is used to estimate uncertainties on the metrics.

    :param models: list of topic labelling models
    :param model_names: list of model names
    :param model_inputs: list of input data for models
    :param Y: ground truth topic labels (one-hot format)
    :param cv: number of folds for cross-validation
    :return: tuple (figure, list of axes)
    """
    n = len(models)
    samples = np.array([sum(Y[cat]) for cat in Y.columns])
    order = np.argsort(samples)[::-1]
    samples = samples[order]
    topics = Y.columns[order]

    def get_cv_scores(scoring, model, X):
        scores = np.array([cross_val_score(model.estimators_[i], X, Y[cat], scoring=scoring, cv=cv)
                           for (i, cat) in enumerate(Y.columns[order])])
        smed = np.median(scores, axis=1)
        smin = np.min(scores, axis=1)
        smax = np.max(scores, axis=1)
        err = np.column_stack([np.abs(smin - smed), np.abs(smax - smed)])
        return [smed, err]

    precision = [get_cv_scores('precision', m, X) for (m, X) in zip(models, model_inputs)]
    recall = [get_cv_scores('recall', m, X) for (m, X) in zip(models, model_inputs)]
    f1 = [get_cv_scores('f1', m, X) for (m, X) in zip(models, model_inputs)]

    # y-axis
    ypos = np.arange(len(samples))

    # figure
    plt.close('all')
    f, (ax1, ax2, ax3, ax4) = plt.subplots(1, 4, sharey=True, figsize=(16, 0.25 * len(samples)))

    # samples subfig
    ax1.set_xlabel('Samples')
    ax1.barh(ypos, samples, align='center',
             color='g', label='Samples', alpha=0.25)
    ax1.set_yticks(ypos)
    ax1.set_yticklabels(topics)
    ax1.invert_yaxis()

    # precision
    ax2.set_xlabel('Precision')
    ax2.set_xlim((-0.05, 1.05))
    for i in range(n):
        (med, err) = precision[i]
        ax2.errorbar(med, ypos, xerr=err.T, color=COLORS[i],
                     fmt='o', capsize=5, label=model_names[i], alpha=0.5)

    # recall
    ax3.set_xlabel('Recall')
    ax3.set_xlim((-0.05, 1.05))
    for i in range(n):
        (med, err) = recall[i]
        ax3.errorbar(med, ypos, xerr=err.T, color=COLORS[i],
                     fmt='o', capsize=5, label=model_names[i], alpha=0.5)

    # f1
    ax4.set_xlabel('F1')
    ax4.set_xlim((-0.05, 1.05))
    for i in range(n):
        (med, err) = f1[i]
        ax4.errorbar(med, ypos, xerr=err.T, color=COLORS[i],
                     fmt='o', capsize=5, label=model_names[i], alpha=0.5)

    ax4.legend(loc='center left', bbox_to_anchor=(1, 1))

    gridlines = []
    for ax in [ax1, ax2, ax3, ax4]:
        ax.grid()
        gridlines += ax.get_xgridlines() + ax.get_ygridlines()
    for line in gridlines:
        line.set_linestyle('-.')

    return (f, (ax1, ax2, ax3, ax4))


def background_composition_pie(Y_true, Y_score, topic, threshold, min_topic_frac=0.05):
    """Create a pie chart illustrating the major background contributions for given label

    Background topics contributing less than *min_topic_frac* will be merged into a
    single contribution called "Other".

    A bar chart is also included illustrating the overall topic composition.

    :param Y_true: ground truth topic labels (one-hot format)
    :param Y_score: topic probability predictions (shape: samples x topics)
    :param topic: name of topic to investigate
    :param threshold: threshold above which to investigate background contributions
    :param min_topic_frac: minimum background sample fraction
    :return: tuple (figure, list of axes)
    """
    ix = Y_true.columns.get_loc(topic)
    y_score = Y_score[:, ix]
    topics = np.array([t for t in Y_true.columns if t != topic])
    composition = np.array([np.sum(Y_true[topic][y_score > threshold]) for topic in topics])

    # combine contributions less than 5%
    tot = np.sum(composition)
    mask = (composition < tot * min_topic_frac)
    other = np.sum(composition[mask])
    topics = np.array(topics[~mask].tolist() + ["Other"])
    composition = np.array(composition[~mask].tolist() + [other])

    # sort
    topics = topics[np.argsort(composition)]
    composition = np.sort(composition)

    # make fig
    fig = plt.figure(figsize=(15, 5))
    # Plot 1: bar
    ax1 = plt.subplot(1, 2, 1)
    ypos = np.arange(len(composition))
    ax1.barh(ypos, composition, align='center')
    ax1.set_yticks(ypos)
    ax1.set_yticklabels(topics)
    ax1.set_xlabel('Samples')

    # Plot 2: pie
    ax2 = plt.subplot(1, 2, 2)
    ax2.pie(composition, labels=topics, autopct='%1.1f%%', startangle=90)
    plt.axis('equal')

    return (fig, (ax1, ax2))


def get_multimodel_sample_size_dependence(models, datasets, labels, sample_fracs, scoring=None, cat_scoring=None):
    """Return performance metrics vs training sample size

    Fractions of data (*sample_fracs*) are randomly sampled from the training dataset
    and used to train the models, which are always evaluated on the full testing datasets.

    :param models: list of topic labelling models
    :param datasets: list of input data for models (each is (training, testing) tuple)
    :param labels: tuple (train, test) of ground truth topic labels (one-hot format)
    :param sample_fracs: list of sample fractions to scan
    :param scoring: sklearn scorer or scoring name for topic averaged metric
    :param cat_scoring: sklearn scorer or scoring name for individual topic metric
    :return: tuple (entries per step, averaged model scores for each step, model scores for each topic for each step)
    """
    # inputs
    (Y_train, Y_test) = labels
    train_size = len(Y_train)
    test_size = len(Y_test)
    train_indices = np.arange(train_size)
    categories = Y_train.columns

    # check input dataset size compatibility
    assert np.all(np.array([X.shape[0] for (X, _) in datasets]) == train_size), \
        "Model training sample sizes are incompatible!"
    assert np.all(np.array([X.shape[0] for (_, X) in datasets]) == test_size), \
        "Model testing sample sizes are incompatible!"

    # values to fill
    entries = []
    scores = [] if scoring is not None else None
    cat_scores = [] if cat_scoring is not None else None
    for frac in sample_fracs:
        # sub-sampling
        subsample_size = int(frac * train_size)
        np.random.seed(42)
        rand_indices = np.random.choice(train_indices, subsample_size, replace=False)
        Y_train_sub = Y_train.iloc[rand_indices]

        # account for active categories (ie have at least 1 True and 1 False label)
        active_cats = [cat for cat in categories if len(Y_train_sub[cat].unique()) == 2]
        if len(active_cats) == 0:
            print("no active categories, skipping frac: ", frac)
            continue
        print("frac: {}, samples: {}, active cats: {}".format(frac, subsample_size, len(active_cats)))
        Y_train_sub = Y_train_sub[active_cats]
        Y_test_sub = Y_test[active_cats]

        # evaluate model
        model_scores = []
        cat_model_scores = []
        for (model, (X_train, X_test)) in zip(models, datasets):
            # print ("evaluating model...")
            X_train_sub = X_train[rand_indices]

            # train
            model_tmp = clone_model(model)
            model_tmp.fit(X_train_sub, Y_train_sub)

            # predict/eval overall
            scorer = Y_test_pred = None
            if scoring is not None:
                scorer = check_scoring(model_tmp, scoring)
                if isinstance(scorer, _PredictScorer):
                    Y_test_pred = model_tmp.predict(X_test)
                elif isinstance(scorer, _ProbaScorer):
                    Y_test_pred = model_tmp.predict_proba(X_test)
                else:
                    assert False, "Scorer not supported"
                model_scores.append(scorer._score_func(Y_test_sub, Y_test_pred, **scorer._kwargs))

            # predict/eval per category
            if cat_scoring is not None:
                cat_scorer = check_scoring(model_tmp.estimators_[0], cat_scoring)
                if scoring is not None and type(scorer) == type(cat_scorer):
                    Y_test_pred_cat = Y_test_pred
                else:
                    if isinstance(cat_scorer, _PredictScorer):
                        Y_test_pred_cat = model_tmp.predict(X_test)
                    elif isinstance(cat_scorer, _ProbaScorer):
                        Y_test_pred_cat = model_tmp.predict_proba(X_test)
                    else:
                        assert False, "Category Scorer not supported"

                # eval
                cat_score = []
                for cat in categories:
                    if cat not in active_cats:
                        s = 0.0
                    else:
                        icat = np.where(Y_test_sub.columns == cat)[0][0]
                        s = cat_scorer._score_func(Y_test_sub[cat],
                                                   Y_test_pred_cat[:, icat],
                                                   **cat_scorer._kwargs)
                    cat_score.append(s)
                cat_model_scores.append(cat_score)

            # Note: this is typically how to call the scorer (but we hacked to avoid multiple prediction)
            # score = scorer(model_tmp, X_test, Y_test_sub)

        entries.append(subsample_size)
        if scoring is not None:
            scores.append(model_scores)
        if cat_scoring is not None:
            cat_scores.append(cat_model_scores)

    entries = np.array(entries)
    if scoring is not None:
        scores = np.array(scores).T
    if cat_scoring is not None:
        cat_scores = np.array(cat_scores).T
    return (entries, scores, cat_scores)


def multimodel_sample_size_dependence_graph(models, model_names, datasets, labels, sample_fracs, scoring=None,
                                            cat_scoring=None):
    """Create graph of performance metric vs training sample size

    Fractions of data (*sample_fracs*) are randomly sampled from the training dataset
    and used to train the models, which are always evaluated on the full testing datasets.

    :param models: list of topic labelling models
    :param model_names: list of model names
    :param datasets: list of input data for models (each is (training, testing) tuple)
    :param labels: tuple (train, test) of ground truth topic labels (one-hot format)
    :param sample_fracs: list of sample fractions to scan
    :param scoring: sklearn scorer or scoring name for topic averaged metric
    :param cat_scoring: sklearn scorer or scoring name for individual topic metric
    :return: tuple (figure, major axis, minor axes (flat list), minor axes (2D list))
    """
    (entries, scores, cat_scores) = get_multimodel_sample_size_dependence(
        models, datasets, labels, sample_fracs, scoring=scoring, cat_scoring=cat_scoring)

    # set figure configuration
    if scoring is None:
        maj_w = maj_h = None
    else:
        maj_w = 3
        maj_h = 2
    if cat_scoring is None:
        ncat = 0
    else:
        ncat = cat_scores.shape[0]

    plt.close('all')
    (f, ax_maj, ax_min, ax_min_ij) = create_awesome_plot_grid(
        ncat, maj_w=maj_w, maj_h=maj_h, min_xlabel="Train sample size", min_ylabel="Score")

    # plot main figure
    if scoring:
        ax = ax_maj
        for j in range(len(models)):
            ax.plot(entries, scores[j], color=COLORS[j], label=model_names[j])
        ax.set_title("Overall", pad=25)
        ax.legend()

    # plot grid with categories
    if cat_scoring:
        # get category sample fractions
        categories = labels[0].columns
        cfracs = np.array([np.sum(labels[0][cat]) / len(labels[0]) for cat in categories])

        # sort categories by size
        order = np.argsort(cfracs)[::-1]
        categories = categories[order]
        cfracs = cfracs[order]
        cat_scores = cat_scores[order]

        # plot subfigs
        for i in range(len(categories)):
            ax = ax_min[i]
            for j in range(len(models)):
                ax.plot(entries, cat_scores[i, j], color=COLORS[j], label=model_names[j])
            pad = 25 if ax.i == 0 else None
            ax.set_title("{} ({:.1f}% frac)".format(categories[i], 100. * cfracs[i]), pad=pad)
        if not scoring: ax_min[0].legend()

    return (f, ax_maj, ax_min, ax_min_ij)


def topic_correlation_matrix(Y):
    """Create MxM correlation matrix for M topics

    Each column represents a given ground truth topic label.
    Each row represents the relative frequency with which other
    ground truth labels co-occur.

    :param Y: ground truth topic labels (one-hot format)
    :return: tuple (figure, axis)
    """
    d = np.array([np.sum(Y[Y[t]], axis=0) / np.sum(Y[t]) for t in Y.columns]) * 100
    d = d.T
    df = pd.DataFrame(d, columns=Y.columns)
    df['topic'] = Y.columns
    df = df.set_index('topic')
    fig, ax = plt.subplots(figsize=(11, 11))
    graph = sns.heatmap(df, annot=True, fmt=".0f", cbar=False, cmap="Blues", linewidths=0.2)
    ax.xaxis.tick_top()  # x axis on top
    ax.xaxis.set_label_position('top')
    ax.set_xlabel('Chosen label')
    ax.set_ylabel('Coincidence of other labels with chosen label [%]')
    _ = plt.xticks(rotation=90)

    return (fig, ax)


def topic_migration_matrix(Y_true, Y_pred):
    """Create MxM migration matrix for M topics

    Each column represents a given ground truth topic label.
    Each row represents the relative frequency with which
    predicted labels are assigned.

    :param Y_true: ground truth topic labels (one-hot format)
    :param Y_pred: topic predictions (one-hot format)
    :return: tuple (figure, axis)
    """
    d = np.array([np.sum(Y_pred[Y_true[t]], axis=0) / np.sum(Y_true[t]) for t in Y_true.columns]) * 100
    d = d.T
    df = pd.DataFrame(d, columns=Y_true.columns)
    df['topic'] = Y_true.columns
    df = df.set_index('topic')
    fig, ax = plt.subplots(figsize=(11, 11))
    graph = sns.heatmap(df, annot=True, fmt=".0f", cbar=False, cmap="Greens", linewidths=0.2)
    ax.xaxis.tick_top()  # x axis on top
    ax.xaxis.set_label_position('top')
    ax.set_xlabel('True label')
    ax.set_ylabel('Frequency of predicted label per true label [%]')
    _ = plt.xticks(rotation=90)

    return (fig, ax)


def false_labels_matrix(Y_true, Y_pred):
    """Create MxM false labels matrix for M topics

    Each column represents a given ground truth topic label.
    Each row represents the absolute number of false predicted
    labels.

    :param Y_true: ground truth topic labels (one-hot format)
    :param Y_pred: topic predictions (one-hot format)
    :return: tuple (figure, axis)
    """
    d = np.array([np.sum(Y_pred[Y_true[t]] & ~Y_true[Y_true[t]], axis=0) for t in Y_true.columns])
    d = d.T
    df = pd.DataFrame(d, columns=Y_true.columns)
    df['topic'] = Y_true.columns
    df = df.set_index('topic')
    fig, ax = plt.subplots(figsize=(11, 11))
    graph = sns.heatmap(df, annot=True, fmt=".0f", cbar=False, cmap="Reds", linewidths=0.2)
    ax.xaxis.tick_top()  # x axis on top
    ax.xaxis.set_label_position('top')
    ax.set_xlabel('True label')
    ax.set_ylabel('Number of false labels for given true label')
    _ = plt.xticks(rotation=90)
    return fig


def keras_train_history_graph(history, metrics):
    """Plot selected performance *metrics* as a function of training epoch.

    :param history: keras training history
    :param metrics: list of metric names to plot
    :return: tuple (figure, list of axes)
    """
    plt.close('all')
    f, axs = plt.subplots(len(metrics), 1, sharex=True, figsize=(8, 8))

    if not isinstance(axs, np.ndarray):
        axs = [axs]

    for (i, metric) in enumerate(metrics):
        ax = axs[i]
        if metric in history:
            x = np.arange(len(history[metric]))
            ax.plot(x, history[metric], c='r', label='train')
        if 'val_' + metric in history:
            x = np.arange(len(history['val_' + metric]))
            ax.plot(x, history['val_' + metric], c='b', label='validation')
        if i == 0: ax.legend()
        ax.grid()
        ax.set_xlabel("epoch")
        ax.set_ylabel(metric)

    return (f, axs)

# EOF


# encoding: utf-8
"""
tune.py
~~~~~~~

Functionality for tuning models

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-05-08"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports

# third party imports
import numpy as np
from matplotlib import pyplot as plt
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import accuracy_score, roc_auc_score

# local imports
from atnlp.core.logger import log, title_break

# globals


def grid_search(X_train, y_train, model, pname, pvals, scoring=None):
    """Perform 1D model hyper-parameter scan using 5-fold cross-validation

    The sklearn grid is returned and a plot of the performance is made.

    :param X_train: training data
    :param y_train: ground truth labels
    :param model: model
    :param pname: model hyperparameter name
    :param pvals: model hyperparameter values
    :param scoring: sklearn performance metric (optional)
    :return: sklearn GridSearchCV
    """
    # configure grid-search parameters
    params = {pname:pvals}

    # run grid-search
    grid = GridSearchCV(model, cv=5, param_grid=params,
                        return_train_score=True,
                        scoring=scoring)
    result = grid.fit(X_train, y_train)

    # plot results
    scan_x = params[pname]
    plt.errorbar(scan_x, grid.cv_results_['mean_test_score'],
                 yerr=grid.cv_results_['std_test_score'],
                 label='test')
    plt.errorbar(scan_x, grid.cv_results_['mean_train_score'],
                 yerr=grid.cv_results_['std_train_score'],
                 label = 'train')
    plt.legend()
    ax = plt.gca()
    ax.grid(True)
    for line in ax.get_xgridlines() + ax.get_ygridlines():
        line.set_linestyle('-.')

    return grid


def find_threshold(y_true, y_score, target_contamination=0.01, show=True):
    """Return binary classification probability threshold that yields contamination closest to target

    :param y_true: ground truth labels
    :param y_score: predicted class scores
    :param target_contamination: target level of contamination (1-precision)
    :param show: make plots
    :return: optimal threshold
    """
    thres_min = 0.0
    thres_max = 1.02
    thres = np.arange(thres_min, thres_max, 0.005)

    # first calculate denominator (total entries passing threshold)
    tots = np.array([np.sum(y_score > v) for v in thres])

    # filter entries where nothing passes
    thres = thres[tots > 0]
    tots = tots[tots > 0]

    # if no points return a valid number, return threshold
    if len(tots) == 0:
        return thres_min

    # calculate contamination
    conts = np.array([np.sum(y_score[y_true == False] > v) for v in thres]) / tots

    # get point closest to target contamination
    idx = (np.abs(conts - target_contamination)).argmin()

    # plot
    if show:
        recall = np.array([np.sum(y_score[y_true == True] > v) for v in thres]) / np.sum(y_true)
        ymin = max(min([target_contamination - 0.1, conts[idx]]), 0.0)
        ymax = min(max([target_contamination + 0.1, conts[idx]]), 1.0)

        plt.figure(figsize=(20, 5))

        # Plot 1: probability distributions
        plt.subplot(1, 3, 1)
        bins = np.arange(0., 1.001, 0.02)
        plt.hist(y_score[y_true == True], bins=bins, label='Signal', alpha=0.5)
        plt.hist(y_score[y_true == False], bins=bins, label='Background', alpha=0.5)
        ax = plt.gca()
        ax.set_yscale('log')
        plt.vlines(x=thres[idx], ymin=ax.get_ylim()[0], ymax=ax.get_ylim()[1], linestyles='--')
        plt.annotate('chosen threshold', xy=(thres[idx], ax.get_ylim()[1] / 10.0),
                     verticalalignment='top', horizontalalignment='right', rotation=90)

        plt.xlabel('threshold')
        plt.legend()

        # Plot 2: threshold tuning
        plt.subplot(1, 3, 2)
        plt.hlines(y=target_contamination, xmin=thres_min, xmax=thres_max, linestyles='--')
        plt.annotate('target', xy=(thres_min, target_contamination),
                     verticalalignment='bottom')
        plt.plot(thres, conts, label='contamination', color='b')
        plt.plot(thres, recall, label='recall', color='r')
        plt.plot(thres[idx], conts[idx], 'bo')
        plt.plot(thres[idx], recall[idx], 'ro')
        # plt.ylim(ymin,ymax)
        plt.legend()

        ax = plt.gca()
        ax.grid(True)
        gridlines = ax.get_xgridlines() + ax.get_ygridlines()
        for line in gridlines:
            line.set_linestyle('-.')
        # ax.set_yscale('log')

        plt.xlabel('threshold')
        plt.ylabel('cont. / recall')

        # Plot 3: ROC
        plt.subplot(1, 3, 3)
        plt.plot(conts, recall)
        plt.xlabel('contamination')
        plt.ylabel('recall')
        plt.plot(conts[idx], recall[idx], 'bo')

        ax = plt.gca()
        ax.grid(True)
        gridlines = ax.get_xgridlines() + ax.get_ygridlines()
        for line in gridlines:
            line.set_linestyle('-.')

        plt.xlim(-0.02, 1.02)
        plt.ylim(-0.02, 1.02)

    return thres[idx]


def fit_xgb_model(alg, X, y, X_test, y_test, useTrainCV=True, cv_folds=5, early_stopping_rounds=50):
    """Fit xgboost model

    :param alg: XGBClassifier (sklearn api class)
    :param X: training data
    :param y: training labels
    :param X_test: testing data
    :param y_test: testing labels
    :param useTrainCV: use cross validation
    :param cv_folds: number of folds for cross-validation
    :param early_stopping_rounds: minimum number of rounds before early stopping
    """
    if useTrainCV:
        import xgboost as xgb
        dtrain = xgb.DMatrix(X, label=y)
        cvresult = xgb.cv(alg.get_xgb_params(), dtrain,
                          num_boost_round=alg.get_params()['n_estimators'],
                          early_stopping_rounds=early_stopping_rounds,
                          nfold=cv_folds, metrics='auc')
        alg.set_params(n_estimators=cvresult.shape[0])

    # Fit the algorithm on the data
    eval_set = [(X, y), (X_test, y_test)]
    alg.fit(X, y, eval_metric='auc', eval_set=eval_set, verbose=False)

    # Predict training set:
    y_pred = alg.predict(X)
    y_prob = alg.predict_proba(X)[:, 1]

    # Print model report:
    title_break("Model Report")
    log().info("Accuracy : %.4g" % accuracy_score(y, y_pred))
    log().info("AUC Score (Train): %f" % roc_auc_score(y, y_prob))

    result = alg.evals_result()
    n = len(result['validation_0']['auc'])

    if useTrainCV:
        x = np.arange(len(cvresult))
        (ytr, eytr) = (cvresult['train-auc-mean'], cvresult['train-auc-std'])
        (yte, eyte) = (cvresult['test-auc-mean'], cvresult['test-auc-std'])

        plt.fill_between(x, ytr - eytr, ytr + eytr, facecolor='r', alpha=0.25, label='train(cv) err')
        plt.fill_between(x, yte - eyte, yte + eyte, facecolor='b', alpha=0.25, label='test(cv) err')
        plt.plot(x, ytr, color='r', linestyle='--', label='train(cv)')
        plt.plot(x, yte, color='b', linestyle='--', label='test(cv)')

    plt.plot(np.arange(n), result['validation_0']['auc'], color='r', label='train')
    plt.plot(np.arange(n), result['validation_1']['auc'], color='b', linewidth=2, label='test')

    plt.legend()


# EOF


# encoding: utf-8
"""
metrics.py
~~~~~~~~~~

Functionality for computing performance metrics. Typically custom metrics not provided by sklearn.

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-05-30"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports 

# third party imports
import numpy as np

# local imports

# globals


def recall_all_score(Y_true, Y_pred):
    """Return the 'recall all' score

    'Recall all' is defined as::

        score := number of examples with all labels correct / number of examples

    :param Y_true: ground truth topic labels (one-hot format)
    :param Y_pred: topic predictions (one-hot format)
    :return: recall all score
    """
    Y_true = Y_true.as_matrix()
    matches = np.all(Y_true == Y_pred, axis=1)
    return np.sum(matches) / len(matches)


def flpd_score(Y_true, Y_pred):
    """Return 'false labels per document' score

    'False labels per document' is defined as::

        score := total number of false labels / number of examples

    :param Y_true: ground truth topic labels (one-hot format)
    :param Y_pred: topic predictions (one-hot format)
    :return: false labels per document score
    """
    Y_true = Y_true.as_matrix()
    return np.sum(Y_pred & ~Y_true) / Y_true.shape[0]


def mlpd_score(Y_true, Y_pred):
    """Missing labels per document score

    'Missing labels per document' is defined as::

        score := total number of missing labels / number of examples

    :param Y_true: ground truth topic labels (one-hot format)
    :param Y_pred: topic predictions (one-hot format)
    :return: missing labels per document score
    """
    Y_true = Y_true.as_matrix()
    return np.sum(~Y_pred & Y_true) / Y_true.shape[0]

# EOF


# encoding: utf-8
"""
parse.py
~~~~~~~~

Functionality for parsing text inputs.

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-05-05"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports
import re
import collections

# third party imports
from sklearn.feature_extraction.stop_words import ENGLISH_STOP_WORDS as STOPWORDS

# local imports

# globals
TOKEN_PATTERN1 = re.compile(r"(?u)\b\w\w+\b")
PAD_WORD = "<pad>"
UNKNOWN_WORD = "<uw>"


# functions
def tokenize1(text):
    """Return tokenized list of strings from raw text input

    :param text: raw text (string)
    :return: list of tokens (strings)
    """
    return TOKEN_PATTERN1.findall(text)


def tokenize_keras(raw_data):
    """Return tokenized list of strings from raw text input using keras functionality

    :param raw_data: raw text (string)
    :return: list of tokens (strings)
    """
    from keras.preprocessing.text import text_to_word_sequence
    return [text_to_word_sequence(d) for d in raw_data]


def filter1(word):
    """Return True if word passes filter

    :param word: string
    :return: True or False
    """
    if not word: return False
    w = word.lower()
    if w in STOPWORDS: return False
    return True


def process_text(text, tokenize=tokenize1, filter=filter1, stem=None, lower=True):
    """Return processed list of words from raw text input

    To be honest, we're currently using sklearn CountVectorizer and
    keras text_to_word_sequence instead of this function.

    :param text: raw text input (string)
    :param tokenize: tokenizing function
    :param filter: filter function
    :param stem: stemming function
    :param lower: convert input text to lowercase if true
    :return: list of strings
    """
    assert tokenize, "Must provide tokenize method for preprocess_text"
    if not text: return []
    if lower: text = text.lower()
    words = tokenize(text)
    if filter: words = [w for w in words if filter(w)]
    if stem: words = [stem(w) for w in words]
    return words


def build_vocab(raw_data, max_size=None):
    """Return dict of word-to-id from raw text data

    If max_size is specified, vocab is truncated to set
    of highest frequency words within size.

    :param raw_data: list of strings
    :param max_size: maximum size of vocab
    :return: word-to-id dict
    """
    data = [w for doc in tokenize_keras(raw_data) for w in doc]
    counter = collections.Counter(data)
    count_pairs = sorted(counter.items(),
                         key=lambda x: (-x[1], x[0]))
    if max_size: count_pairs = count_pairs[:max_size]
    words, _ = list(zip(*count_pairs))
    word_to_id = dict(zip(words, range(len(words))))
    word_to_id[UNKNOWN_WORD] = len(word_to_id)
    word_to_id[PAD_WORD] = len(word_to_id)
    return word_to_id


def raw_to_ids(raw_data, word_to_id):
    """Convert raw text data into integer ids

    :param raw_data: raw text data (list of strings)
    :param word_to_id: word-to-id dict
    :return: list of list of integer ids
    """
    docs = tokenize_keras(raw_data)
    uid = word_to_id[UNKNOWN_WORD]
    return [[word_to_id.get(w, uid) for w in doc] for doc in docs]


# EOF


# encoding: utf-8
"""
grid.py
~~~~~~~~~~~~~~~

Functionality for creating hyperparameter grid scans

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-06-01"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports
import os
import stat
from datetime import datetime

# third party imports
from sklearn.model_selection import ParameterGrid

# local imports

# globals


def create(params):
    """Return parameter grid (cross-product of individual parameter scans)

    The scan is defined by the *params* dictionary, where parameters to
     scan should be given a list of parameter values. Ie it should be in format:

    .. code-block:: python

        {'par1': [val1, val2],
         'par2': [val1, val2, val3],
         'par3': [val1]
         ...
         }

    :param params: params dictionary (with lists for parameters to be scanned)
    :return: model params for each scan point (list of dicts)
    """
    for (k,v) in params.items():
        if not isinstance(v, list):
            params[k] = [v]
    return list(ParameterGrid(params))


def tofile(grid, counter, exec, filename='grid.sh', shutdown=True):
    """Write parameter grid as sequence of commands in bash script

    *exec* is the command line function

    *counter* is a flag passed to the command line function that
    receives the command index as a string format argument. It can
    be used to increment the output model number, eg `-o model.{}.h5`

    If *shutdown* is true, a shutdown command will be included
    at the end of the script (useful for shutting down virtual
    instances after completing scan eg. in google cloud).

    :param grid: parameter grid
    :param counter: command line increment flag (string)
    :param exec: command line exec (string)
    :param filename: output file name
    :param shutdown: include shutdown command
    """
    with open(filename, 'w') as f:
        f.write("#!/bin/bash\n")
        f.write("#\n")
        f.write("# Grid scan created at {}\n".format(datetime.now()))
        f.write("#\n")
        f.write("# Suggest to execute with: \n")
        f.write("#\n")
        f.write("#   nohup ./{} &\n".format(filename))
        f.write("#\n")
        for (i,args) in enumerate(grid):
            count_str = counter.format(i)
            arg_str = " ".join(["--{}".format(k) if v is True
                                else "--{} {}".format(k,v)
                                for (k,v) in args.items()
                                if v is not False])
            f.write("{} {} {}\n".format(exec, count_str, arg_str))
        if shutdown:
            f.write("sudo poweroff\n")

    # make exec
    os.chmod(filename, os.stat(filename).st_mode | stat.S_IEXEC)


# EOF


# encoding: utf-8
"""
setup.py
~~~~~~~~

Function to setup the atnlp framework (call it before using the framework).

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-05-29"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports 
import logging
import warnings
from time import localtime, asctime
import os

# third party imports
from sklearn.exceptions import UndefinedMetricWarning

# local imports
from atnlp.core.logger import log

# globals


def setup(
        log_level=None,
        suppress_warnings=None,
        tf_loglvl=2,
        batch_mode=False,
):
    """Global setup for atnlp framework

    :param log_level: logging output level
    :type log_level: logging.LEVEL (eg. DEBUG, INFO, WARNING...)
    :param suppress_warnings: list of warnings to suppress
    :type suppress_warnings: bool
    :param tf_loglvl: tensorflow log level
    :type tf_loglvl: int
    """
    # log level
    if log_level == None: log_level = logging.INFO
    log().setLevel(log_level)

    # get this show rolling
    log().info("Starting Job: %s" % (asctime(localtime())))

    # Suppress warnings
    if suppress_warnings is None:
        suppress_warnings = [
            UndefinedMetricWarning,
            UserWarning,
        ]
    log().info("Suppressing following warnings:")
    for warn in suppress_warnings:
        log().info("    {}".format(warn))
        warnings.filterwarnings("ignore", category=warn)

    # Set tensorflow log level
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = str(tf_loglvl)

    # set non-interactive backend if display not available
    if batch_mode or (os.name == 'posix' and "DISPLAY" not in os.environ):
        #import matplotlib
        #matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        plt.switch_backend('agg')

# EOF


# encoding: utf-8
"""
wordmatch.py
~~~~~~~~~~~~

Pipeline converting raw text to sparse bag-of-words representation and feeding to
a custom key-word based topic labelling model.

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-06-07"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# imports
from sklearn.pipeline import Pipeline
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.multiclass import OneVsRestClassifier
from atnlp.model.wordmatch import WordMatchClassifier

model = Pipeline([
    ('bow', CountVectorizer()),
    ('wmc', OneVsRestClassifier(WordMatchClassifier())),
    ])

# EOF


# encoding: utf-8
"""
table.py
~~~~~~~~

Functionality for creating performance summary tables.

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-05-05"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports

# third party imports
import numpy as np
import pandas as pd
from sklearn.metrics import precision_recall_fscore_support

# local imports
from atnlp.eval.metrics import recall_all_score, flpd_score, mlpd_score

# globals


def topic_labelling_summary_table(Y_true, Y_pred, sample_min=None, thresholds=None):
    """Return topic labelling summary table for single model predictions

    Contents of the table includes the following entries per topic:

    - samples: total number of examples
    - standard metrics: precision, recall, f1
    - fl: total number of false labels (for topic)
    - flps: false labels for topic / topic samples
    - flpd: false labels for topic / total documents
    - ml: total numebr of missing labels (for topic)
    - mlps: missing labels for topic / topic samples
    - mlpd: missing labels for topic / total documents

    If *sample_min* is specified, topics with fewer examples will be omitted.

    *thresholds* is a list of one threshold per category, which if specified,
    will be applied to *Y_pred* to generate class predictions. In this case
    *Y_pred* is assumed to be a matrix of class probability scores rather than
    predictions.

    :param Y_true: ground truth topic labels (one-hot format)
    :param Y_pred: topic predictions (one-hot format)
    :param sample_min: minimum number of examples per topic
    :param thresholds: list of thresholds per category (optional)
    :return: summary table (pandas DataFrame)
    """
    # filter topics with too few samples    
    if sample_min is not None:
        samples = np.array([sum(Y_true[topic]) for topic in Y_true.columns])
        Y_true = Y_true[Y_true.columns[samples > sample_min]]
        Y_pred = Y_pred[:, samples > sample_min]
        if thresholds is not None:
            thresholds = thresholds[samples > sample_min]

    # define column order
    columns = ['topic', 'samples',
               'precision', 'recall', 'f1',
               'fl', 'flps', 'flpd',
               'ml', 'mlps', 'mlpd']

    # build table data
    scores = np.array([np.array(precision_recall_fscore_support(Y_true[c], Y_pred[:, i]))[:3,1]
                       for (i,c) in enumerate(Y_true.columns)])
    samples = np.sum(Y_true, axis=0)
    fl = np.sum(Y_pred & ~Y_true, axis=0)
    flps = fl / samples
    flpd = fl / len(Y_true)
    ml = np.sum(~Y_pred & Y_true, axis=0)
    mlps = ml / samples
    mlpd = ml / len(Y_true)

    data = {'topic' :Y_true.columns, 'samples': samples,
        'precision': scores[:,0], 'recall': scores[:,1], 'f1': scores[:,2],
        'fl': fl, 'flps':flps, 'flpd':flpd,
        'ml': ml, 'mlps':mlps, 'mlpd':mlpd}

    if thresholds is not None:
        data['threshold'] = thresholds
        columns += ['threshold']

    # return table sorted by number of samples
    return pd.DataFrame(data, columns=columns) \
        .sort_values(by='samples', ascending=False) \
        .round({k:3 for k in columns if k not in ['topic', 'samples', 'fl', 'ml', 'threshold']})


def multimodel_topic_labelling_summary_tables(Y_true, Y_preds, model_names, sample_min=None, thresholds=None):
    """Return dictionary of topic labelling summary tables for multiple model predictions

    The dictionary includes a single table for each of the metrics included in
    :func:`topic_labelling_summary_table`, where the key is the metric name.

    An overall summary table (with key *summary*) is also provided, including the following metrics:

    - pre_mic, rec_mic, f1_mic: precision, recall and f1 scores using 'micro' averaging over topics
    - recall_all: recall calculated requiring all labels in document correct (see :func:`atnlp.eval.metrics.recall_all_score`)
    - flpd, mlpd: false/missing labels per document (see :func:`atnlp.eval.metrics.flpd_score`, :func:`atnlp.eval.metrics.mlpd_score`)

    In each table, metrics are provided for each of the models provided.

    If *sample_min* is specified, topics with fewer examples will be omitted.

    *thresholds* is a list of one threshold per category per model, which if specified,
    will be applied to *Y_pred* to generate class predictions. In this case
    *Y_pred* is assumed to be a matrix of class probability scores rather than
    predictions.

    :param Y_true: ground truth topic labels (one-hot format)
    :param Y_preds: list of topic predictions for each model (one-hot format)
    :param model_names: name of each model
    :param sample_min: minimum number of examples per topic
    :param thresholds: list of thresholds per category (optional)
    :return: dict of summary tables (pandas DataFrames)
    """
    assert len(model_names) == len(Y_preds), "model_names must be same length as Y_preds!"
    if thresholds: assert len(thresholds) == len(Y_preds)
    else:          thresholds = [None] * len(Y_preds)
    n = len(model_names)
    tables = dict()

    # get per-model tables
    model_tables = [topic_labelling_summary_table(Y_true, Y_preds[i],
                                                  sample_min=sample_min,
                                                  thresholds=thresholds[i])
                    for i in range(n)]

    # create per-topic summary tables for each metric
    topic = model_tables[0]['topic']
    samples = model_tables[0]['samples']
    columns = [c for c in model_tables[0].columns if c not in ['topic','samples','threshold']]
    for c in columns:
        df = pd.DataFrame({'topic':topic, 'samples':samples}, columns=['topic','samples']+model_names)
        for (i,n) in enumerate(model_names):
            df[n] = model_tables[i][c]
        tables[c] = df

    # create overall metric table
    scores = np.array([np.array(precision_recall_fscore_support(Y_true, Y_pred))[:3,1]
                       for Y_pred in Y_preds])

    columns = ['model', 'pre_mic', 'rec_mic', 'f1_mic', 'rec_all', 'flpd', 'mlpd']
    ave_table = pd.DataFrame({
            'model': model_names,
            'pre_mic': scores[:,0], 'rec_mic': scores[:,1], 'f1_mic': scores[:,2],
            'rec_all': [recall_all_score(Y_true, Y_pred) for Y_pred in Y_preds],
            'flpd': [flpd_score(Y_true, Y_pred) for Y_pred in Y_preds],
            'mlpd': [mlpd_score(Y_true, Y_pred) for Y_pred in Y_preds],
            },
            columns = columns) \
        .sort_values(by='f1_mic', ascending=False) \
        .round({k:3 for k in columns if k not in ['model']})

    tables['summary'] = ave_table
    return tables


# EOF



# encoding: utf-8
"""
io.py
~~~~~

Functionality for reading and writing datasets

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-06-07"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports 

# third party imports
import pandas as pd
import numpy as np

# local imports

# globals


def write_raw(X, filename):
    """Write raw text data to file

    :param X: list of strings
    :param filename: name of output file
    """
    with open(filename, 'w') as f:
        for l in X:
            l = repr(l.encode('utf-8'))
            f.write(l)
            f.write('\n')


def read_raw(filename):
    """Read raw text data from file

    :param filename: name of input file
    :return: list of strings
    """
    with open(filename, 'r') as f:
        X = f.readlines()
    return X


def write_one_hot_labels(Y, filename):
    """Write topic labels to file in one-hot form

    :param Y: topic labels (one-hot DataFrame, M x N)
    :param filename: name of output file
    """
    if not isinstance(Y, pd.DataFrame):
        df = pd.DataFrame(Y)
        df.to_csv(filename, index=False, header=False)
    else:
        Y.to_csv(filename, index=False)


def read_one_hot_labels(filename):
    """Read topic labels from file in one-hot form

    :param filename: name of input file
    :return: topic labels (one-hot DataFrame, M x N)
    """
    return pd.read_csv(filename, dtype=np.bool)


# EOF


# encoding: utf-8
"""
svm.py
~~~~~~

Pipeline converting raw text to sparse tfidf representation and feeding to
a support vector machine for topic labelling.

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-06-07"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# imports
from sklearn.pipeline import Pipeline
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.multiclass import OneVsRestClassifier
from sklearn.svm import SVC

# model
model = Pipeline([
    ('tfidf', TfidfVectorizer()),
    ('svc', OneVsRestClassifier(SVC())),
    ])

# EOF


#!/usr/bin/env python
# encoding: utf-8
"""
print_modules.py
~~~~~~~~~~~~~~~~

Print list of modules in atnlp (used for auto sphinx documentation)

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-06-13"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# local imports
from atnlp.core.helpers import get_modules


def main():
    for m in get_modules('atnlp'):
        print(".. automodule:: {}\n    :members:".format(m))


if __name__ == "__main__":
    main()
# EOF


# encoding: utf-8
"""
wordmatch.py
~~~~~~~~~~~~

Implements key-word based topic labelling classifier

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-06-07"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports 
import types

# third party imports
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, ClassifierMixin

# local imports
from atnlp.core.logger import log, title_break

# globals


class WordMatchClassifier(BaseEstimator, ClassifierMixin):
    """Keyword based binary topic classifier

    The classifier works by assigning the positive class to
    any example that contains any of the given keywords.
    It is assumed the input data is in the bag-of-words format.

    The set of keywords is fit to the training data using the
    following algorithm. Initially, a single keyword that
    maximises $rp^4$ is selected, where $r$ is recall and $p$
    is precision. The choice $p^4$ is made to strongly penalise
    false positives. This process is repeated, adding keywords
    until the metric decreases.

    TODO: allow predefinition of keywords, that would then be built on?
    TODO: include max number of words as hyperparameter

    :param df_threshold: minimum document frequency for keywords
    """
    def __init__(self, df_threshold=0.1):
        self.df_threshold = df_threshold

        self._keywords = None

    def fit(self, X, y):
        """Fit model to data

        :param X: data (bag-of-words format)
        :param y: binary classification labels
        """
        def next_best_word(X_sig_, X_bkg_, words):
            # get decision from existing words
            mask_sig = self._word_based_predict(X_sig_, words)
            mask_bkg = self._word_based_predict(X_bkg_, words)

            # calculate tp and fp from existing mask
            tp_base = np.sum(mask_sig)
            fp_base = np.sum(mask_bkg)

            # calculate tp/fn/fp arrays for each word (accounting for mask contributions)
            tp = np.sum(X_sig_[~mask_sig] > 0, axis=0) + tp_base
            fn = np.sum(X_sig_[~mask_sig] == 0, axis=0)
            fp = np.sum(X_bkg_[~mask_bkg] > 0, axis=0) + fp_base
            recall = tp / (tp + fn)
            precision = tp / (tp + fp)
            score = recall * np.power(precision, 4)

            # select best scorer
            i = np.argsort(score)[::-1][0]
            return (i, score[i])


        # convert sparse-matrix from BOW into standard matrix
        # (probably inefficient, but the sparse matrix exhibits
        #  a lot of non-intuitive behavior that ultimately leads
        #  to bugs)
        if hasattr(X, 'toarray') and isinstance(X.toarray, types.MethodType):
            X = X.toarray()

        # signal/background datasets
        X_sig = X[y.astype(np.bool)]
        X_bkg = X[~y.astype(np.bool)]

        # select words with document frequency above threshold
        word_mask = np.sum(X_sig > 0, axis=0) > (self.df_threshold * np.sum(y))
        X_sig = X_sig[:, word_mask]
        X_bkg = X_bkg[:, word_mask]

        # iteratively add keywords until optimisation score decreases
        keyword_indices_in_mask = []
        score = None
        while True:
            (w, s) = next_best_word(X_sig, X_bkg, keyword_indices_in_mask)
            if score is None or s > score:
                keyword_indices_in_mask.append(w)
                score = s
            else:
                break

        self._keywords = list(np.arange(X.shape[1])[word_mask][keyword_indices_in_mask])

    def predict(self, X):
        """

        :param X:
        :return:
        """
        if hasattr(X, 'toarray') and isinstance(X.toarray, types.MethodType):
            X = X.toarray()

        return self._word_based_predict(X, self._keywords)

    def decision_function(self, X):
        return self.predict(X)

    def _word_based_predict(self, X, keyword_indices):
        """Return single-class membership prediction based on `keyword_indices`
        (if match to any keyword then return True, otherwise False)

        Will return all false if `keyword_indices` is empty

        :param X: BOW matrix
        :param keyword_indices: indices of keywords in X
        :return: np.array (size `n` where n is rows in X)
        """
        # stack predictions for each word in rows
        # (zeros row in case keyword_indices is empty)
        y_per_word = np.vstack(
            [np.zeros(X.shape[0])] + [X[:, i] > 0 for i in keyword_indices]
        )

        # build logical or from rows
        return np.any(y_per_word, axis=0)


def display_keywords(model, topic_names, vocab):
    """Print keywords for WordMatchClassifier instances in OneVsRestClassifier

    :param model: OneVsRestClassifier containing WordMatchClassifier instances
    :param topic_names: topic for each model instance in OneVsRest
    :param vocab: id-to-word dictionary for bag-of-words input data
    """
    title_break('Topic Keywords')
    for (i, t) in enumerate(topic_names):
        keywords = vocab[model.estimators_[i]._keywords]
        log().info("{:20s}: {}".format(t, keywords))


def get_keyword_dataframe(model, topic_names, vocab):
    """Return pandas DataFrame with keywords for WordMatchClassifier instances in OneVsRestClassifier

    :param model: OneVsRestClassifier containing WordMatchClassifier instances
    :param topic_names: topic for each model instance in OneVsRest
    :param vocab: id-to-word dictionary for bag-of-words input data
    :return: pandas DataFrame
    """
    return pd.DataFrame({
        'topic': topic_names,
        'keywords': [', '.join(vocab[eval._keywords]) for eval in model.estimators_]
    })[['topic', 'keywords']]


# EOF


from setuptools import find_packages, setup
from glob import iglob
import os

# paths
TOPDIR = os.path.dirname(__file__)
ATNLP = os.path.join(TOPDIR,'atnlp')
SCRIPTS = os.path.join(TOPDIR, 'scripts')
SHARE = os.path.join(ATNLP, 'share')

# scripts
script_exts = ['*.py', '*.sh']
scripts = [p for ext in script_exts for p in
           iglob(os.path.join(SCRIPTS,'**',ext), recursive=True)]
# data
data_exts = ['*.py', '*.yml']
data = [os.path.relpath(f, ATNLP) for ext in data_exts
        for f in iglob(os.path.join(SHARE,'**',ext),recursive=True)]

# requirements
install_requires=[
    "matplotlib",
    "nltk",
    "numpy",
    "pandas",
    "pyyaml",
    "scikit-learn",
    "scipy",
    "seaborn",
    "xgboost",
]
extras_requires={
    'deep': ['gensim', 'keras', 'tensorflow'],
    'book': ['ipykernel', 'ipython', 'jupyter', 'notebook'],
    'docs': ['sphinx', 'sphinx_rtd_theme'],
    'dev': ['bumpversion']
}
extras_requires['extras'] = list(set([a for b in extras_requires.values() for a in b]))

setup(
    name='atnlp',
    description='A tool for natural language processing in python based on scipy and sklearn',
    long_description=open(os.path.join(TOPDIR, 'README.md')).read(),
    long_description_content_type='text/markdown',
    url='https://github.com/wedavey/atnlp',
    author='Will Davey',
    author_email='wedavey@gmail.com',
    license='MIT',
    packages=find_packages(),
    package_data={'atnlp':data},
    python_requires=">=3.5",
    install_requires=install_requires,
    extras_require=extras_requires,
    scripts=scripts,
    version='0.0.9',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



#!/usr/bin/env python
# encoding: utf-8
"""
install_reuters.py
~~~~~~~~~~~~~~~~~~

Install the reuters corpus using the nltk api.

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-05-30"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports 
import os

# third party imports
from nltk.downloader import download

# local imports
from atnlp import NLTK_DIR

# globals
def main():
    if not os.path.exists(NLTK_DIR):
        os.makedirs(NLTK_DIR)
        
    download('reuters', download_dir=NLTK_DIR)


if __name__ == '__main__':
    main()
# EOF



#!/usr/bin/env python
# encoding: utf-8
"""
reuters_to_txt.py
~~~~~~~~~~~~~~~~~

Convert installed reuters corpus into standardised text format

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-06-06"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports 
from argparse import ArgumentParser
import logging

# third party imports

# local imports
from atnlp.core.setup import setup
from atnlp.core.logger import log
from atnlp.data.io import write_raw, write_one_hot_labels
from atnlp.data.reuters import get_topics, get_labels, get_data

# globals


def main():
    # parse args
    description = "Convert Reuters dataset to standard text format"
    parser = ArgumentParser(description=description)
    parser.add_argument('-v', '--verbose', action='store_true',
                        help="Set logging level to DEBUG")
    parser.add_argument('--min-samples', type=int, default=100,
                        help="Minimum number of samples per category [default: 100]")
    parser.add_argument('--topics', help="comma separated list of topics")
    args = parser.parse_args()

    # setup atnlp framework
    log_level = logging.DEBUG if args.verbose else None
    setup(log_level=log_level)

    # select topics
    if args.topics:
        topics = args.topics.split(',')
    else:
        topics = get_topics(min_samples=args.min_samples)
    log().info("{} topics selected.".format(len(topics)))

    # get topic labels (MxN data frame of bools: M categories, N documents)
    log().info("getting topic labels...")
    (Y_train, Y_test) = get_labels(topics)
    log().info("Writing to labels_train.txt")
    write_one_hot_labels(Y_train, 'labels_train.txt')
    log().info("Writing to labels_test.txt")
    write_one_hot_labels(Y_test, 'labels_test.txt')

    # get data iterators
    # Note: we also use test data because model currently requires
    #       vocab from all samples to get be predictions
    log().info("getting topic data...")
    (X_train, X_test) = get_data(topics)
    log().info("Writing to data_train.txt")
    write_raw(X_train, "data_train.txt")
    log().info("Writing to data_test.txt")
    write_raw(X_test, "data_test.txt")


if __name__ == '__main__':
    main()

# EOF


#!/usr/bin/env python
# encoding: utf-8
"""
install_glove.py
~~~~~~~~~~~~~~~~

Install pretrained glove word embeddings

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-05-30"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports 
import os
import urllib
import sys
import hashlib
import zipfile

# third party imports

# local imports
from atnlp.model.embed import EMB_DIR

# globals
URL = "http://nlp.stanford.edu/data/glove.6B.zip"
MD5ZIP = "056ea991adb4740ac6bf1b6d9b50408b"
MD5TXT = "b78f53fb56ec1ce9edc367d2e6186ba4"
MD5W2V = "0833e974ac24308f448031cd1d0e3c76"
TARGETZIP = os.path.join(EMB_DIR, 'glove.6B.zip')
TARGETTXT = os.path.join(EMB_DIR, 'glove.6B.300d.txt')
TARGETW2V = os.path.join(EMB_DIR, 'glove.6B.300d.w2vformat.txt')

def main():
    # check if already installed
    if os.path.exists(TARGETW2V):
        if md5hash(TARGETW2V) == MD5W2V:
            print("glove embeddings already installed at: {}".format(TARGETW2V))
            return
        else:
            print("glove embbeddings found but appear corrupt, removing...")
            os.remove(TARGETW2V)

    # prepare directory
    if not os.path.exists(EMB_DIR):
        os.makedirs(EMB_DIR)
    
    # check if txt already available else install 
    if not os.path.exists(TARGETTXT) or md5hash(TARGETTXT) != MD5TXT: 

        # check for zip
        if os.path.exists(TARGETZIP) and md5hash(TARGETZIP) != MD5ZIP:
            print("glove embbeddings zip found but appers corrupt, removing...")
            os.remove(TARGETZIP)

        # download zip
        if not os.path.exists(TARGETZIP):
            print("downloading glove embeddings from {}".format(URL))
            urllib.request.urlretrieve(URL, TARGETZIP, reporthook)

        # extract zip
        print("extracting zip...")
        with zipfile.ZipFile(TARGETZIP,"r") as zip_ref:
            zip_ref.extractall(EMB_DIR)

        # remove zip 
        print("cleaning up zip...")
        os.remove(TARGETZIP)

    # convert 
    print("converting to w2v format...")
    from gensim.scripts.glove2word2vec import glove2word2vec
    glove2word2vec(TARGETTXT, TARGETW2V)


def reporthook(blocknum, blocksize, totalsize):
    readsofar = blocknum * blocksize
    if totalsize > 0:
        percent = readsofar * 1e2 / totalsize
        s = "\r%5.1f%% %*d / %d" % (
            percent, len(str(totalsize)), readsofar, totalsize)
        sys.stderr.write(s)
        if readsofar >= totalsize: # near the end
            sys.stderr.write("\n")
    else: # total size is unknown
        sys.stderr.write("read %d\n" % (readsofar,))

def md5hash(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


if __name__ == "__main__":
    main()
# EOF



# encoding: utf-8
"""
reuters.py
~~~~~~~~~~

Functionality to read in Reuters corpus using the nltk module

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-05-05"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports

# third party imports
from nltk.corpus import reuters
import numpy as np
import pandas as pd
from sklearn.preprocessing import MultiLabelBinarizer

# local imports

# globals


def get_data(cats=None, tokenize=None):
    """Return raw text data from Reuters corpus in (train, test) tuple

    If *cats* is specified, data is filtered to only contain documents
    from the specified categories.

    If *tokenize* is specified, data is tokenized.

    :param cats: categories
    :param tokenize: tokenization function
    :return: tuple of (train, test) data (each is list of strings)
    """
    return (get_data_train(cats,tokenize=tokenize), get_data_test(cats,tokenize=tokenize))


def get_labels(cats=None):
    """Return topic labels (one-hot format) from Reuters corpus in (train, test) tuple

    :param cats: categories
    :return: tuple of (train, test) topic labels (one-hot format)
    """
    return (get_labels_train(cats), get_labels_test(cats))


def get_data_train(cats=None, tokenize=None):
    """Return raw text training data (cf *get_data*)

    :param cats: categories
    :param tokenize: tokenization function
    :return: train data (list of strings)
    """
    return ReutersIter(train_filenames(cats=cats), tokenize=tokenize)


def get_data_test(cats=None, tokenize=None):
    """Return raw text testing data (cf *get_data*)

    :param cats: categories
    :param tokenize: tokenization function
    :return: test data (list of strings)
    """
    return ReutersIter(test_filenames(cats=cats), tokenize=tokenize)


def get_labels_train(cats=None):
    """Return training set topic labels (one-hot format) from Reuters corpus (cf *get_labels*)

    :param cats: categories
    :return: train topic labels (one-hot format)
    """
    return labels(train_filenames(cats), cats=cats)


def get_labels_test(cats=None):
    """Return testing set topic labels (one-hot format) from Reuters corpus (cf *get_labels*)

    :param cats: categories
    :return: test topic labels (one-hot format)
    """
    return labels(test_filenames(cats), cats=cats)


def get_topics(min_samples=None):
    """Return set of topics from Reuters corpus

    If *min_samples* is specified, only topics with at
    least that many examples are included.

    :param min_samples: minimum number of example per topic
    :return: list of topics
    """
    cats = reuters.categories()
    if min_samples is not None:
        cats = [c for c in reuters.categories() if len(reuters.fileids(c)) >= min_samples]
    return cats


def train_filenames(cats=None):
    """Return filenames of training examples

    If *cats* is specified, filenames are filtered to only contain documents
    from the specified categories.

    :param cats: categories
    :return: list of filenames
    """
    return np.array([f for f in reuters.fileids(cats) if f.startswith('train')])


def test_filenames(cats=None):
    """Return filenames of testing examples

    If *cats* is specified, filenames are filtered to only contain documents
    from the specified categories.

    :param cats: categories
    :return: list of filenames
    """
    return np.array([f for f in reuters.fileids(cats) if f.startswith('test')])


def labels(filenames, cats=None):
    """Return topic labels (one-hot format) for given files

    :param filenames: selected files from Reuters dataset
    :param cats: categories to filter (optional)
    :return: topic labels (one-hot format)
    """
    if cats is None: cats = reuters.categories()
    data = [[c for c in reuters.categories(f) if c in cats] for f in filenames]
    mb = MultiLabelBinarizer(classes = cats)
    onehot = mb.fit_transform(data)
    df = pd.DataFrame(onehot, columns=cats)
    return df


class ReutersIter(object):
    """Reuters dataset iterator

    Implements generator instead of reading full dataset into memory.
    However, its not super necessary coz this dataset is small, and
    most of the time we actually create a list from this anyway.

    :param files: list of files to iterate over
    :param tokenize: tokenization function (optional)
    """
    def __init__(self, files, tokenize=None):
        self.files = files
        self.tokenize = tokenize

    def __iter__(self):
        for i in range(len(self.files)):
            yield self[i]

    def __getitem__(self, key):
        data = reuters.raw(self.files[key])
        if self.tokenize is None:
            return data
        return self.tokenize(data)


# EOF


# encoding: utf-8
"""
logger.py
~~~~~~~~~

Framework logging functionality

"""
__author__ = "Will Davey"
__email__ = "wedavey@gmail.com"
__created__ = "2018-05-29"
__copyright__ = "Copyright 2018 Will Davey"
__license__ = "MIT https://opensource.org/licenses/MIT"

# standard imports 
import logging
import sys, os

# third party imports

# local imports

# globals
#: is global logger initialized (global variable)
gInitialized = False

# asci color codes
DEFAULT  = u'\x1b[39;49m'
BLUE     = u'\x1b[34m'
BLUEBOLD = u'\x1b[1;34m'
RED      = u'\x1b[31m'
REDBOLD  = u'\x1b[1;31m'
REDBKG   = u'\x1b[1;41;37m'
YELLOW   = u'\x1b[33m'
UNSET    = u'\x1b[0m'


def initialize(level=None):
    """initialize global logger

    :param level: logging output level
    :type level: logging.LEVEL (eg. DEBUG, INFO, WARNING...)
    """
    logging.basicConfig(
        filemode='w',
        level=level if level != None else logging.INFO,
        format='[%(asctime)s %(levelname)-7s]  %(message)s',
        datefmt='%H:%M:%S',
    )
    logging.getLogger("global")
    if supports_color():
        logging.StreamHandler.emit = add_coloring_to_emit_ansi(logging.StreamHandler.emit)


def log():
    """Return global logger"""
    global gInitialized
    if not gInitialized:
        initialize()
        gInitialized = True
    return logging.getLogger("global")


def setLevel(level):
    """Set global logging level"""
    log().setLevel(level)


def supports_color():
    """
    Returns True if the running system's terminal supports color, and False
    otherwise.
    """
    plat = sys.platform
    supported_platform = plat != 'Pocket PC' and (plat != 'win32' or 'ANSICON' in os.environ)
    # isatty is not always implemented
    is_a_tty = hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()
    if not supported_platform or not is_a_tty:
        return False
    return True


def add_coloring_to_emit_ansi(fn):
    # add methods we need to the class
    def new(*args):
        levelno = args[1].levelno
        if (levelno >= logging.CRITICAL):
            color = REDBKG
        elif (levelno >= logging.ERROR):
            color = REDBOLD
        elif (levelno >= logging.WARNING):
            color = RED
        elif (levelno >= logging.INFO):
            color = DEFAULT
        elif (levelno >= logging.DEBUG):
            color = YELLOW
        else:
            color = YELLOW
        args[1].msg = color + args[1].msg + UNSET
        return fn(*args)

    return new


def section_break(title=None):
    """Print section break to logger (at INFO level)

    :param title: section title (optional)
    """
    log().info("")
    log().info("="*60)
    if title:
        log().info(title)
        log().info("-"*len(title))
        log().info("")

def title_break(title):
    """Print title to logger (at INFO level)"""
    log().info("")
    log().info("-" * len(title))
    log().info(title)
    log().info("-" * len(title))
    log().info("")


# EOF